from typing import List

from cement import Controller, ex

from banyan.api.role import RoleAPI
from banyan.controllers.base import Base
from banyan.model.role import RoleInfo, Role


class RoleController(Controller):
    class Meta:
        label = 'role'
        stacked_type = 'nested'
        stacked_on = 'base'
        help = 'manage user and workload roles'

    @property
    def _client(self) -> RoleAPI:
        return self.app.client.roles

    @ex(help='list roles')
    def list(self):
        roles: List[RoleInfo] = self._client.list()
        results = list()
        headers = ['Name', 'ID', 'Created', 'Last Updated']
        for role in roles:
            new_row = [role.role_name, role.role_id,
                       role.created_at.strftime(Base.TABLE_DATE_FORMAT),
                       role.last_updated_at.strftime(Base.TABLE_DATE_FORMAT)]
            results.append(new_row)
        results.sort(key=lambda x: x[0])
        self.app.render(results, handler='tabulate', headers=headers, tablefmt='simple')

    @ex(help='show the definition of a role',
        arguments=[
            (['role_name'],
             {
                 'metavar': 'role_name_or_id',
                 'help': 'Name or ID of the role to display.'
             }),
        ])
    def get(self):
        info: RoleInfo = self._client[self.app.pargs.role_name]
        role_json = Role.Schema().dump(info.role)
        # colorized_json = highlight(policy_json, lexers.JsonLexer(), formatters.Terminal256Formatter(style="default"))
        self.app.render(role_json, handler='json', indent=2, sort_keys=True)

    @ex(help='create a new role from a JSON specification',
        arguments=[
            (['role_spec'],
             {
                 'help': 'JSON blob describing the new role to be created, or a filename '
                         'containing JSON prefixed by "@" (example: @service.json).'
             }),
        ])
    def create(self):
        spec = Base.get_json_input(self.app.pargs.role_spec)
        role = Role.Schema().load(spec)
        role_id = self._client.create(role)
        self.app.print(f'Role ID: {role_id}')

    @ex(help='update an existing role',
        arguments=[
            (['policy_spec'],
             {
                 'help': 'JSON blob describing the role to be updated, or a filename '
                         'containing JSON prefixed by "@" (example: @service.json).'
             }),
        ])
    def update(self):
        policy_spec = Base.get_json_input(self.app.pargs.role_spec)
        policy = Role.Schema().load(policy_spec)
        policy_info = self._client.update(policy)
        self.app.render(RoleInfo.Schema().dump(policy_info), handler='json', indent=2, sort_keys=True)

    @ex(help='delete a role',
        arguments=[
            (['role_name'],
             {
                 'metavar': 'role_name_or_id',
                 'help': 'Name or ID of the role to delete.'
             }),
        ])
    def delete(self):
        role: RoleInfo = self._client[self.app.pargs.role_name]
        self.app.print(self._client.delete(role))

    @ex(help='enable a role',
        arguments=[
            (['role_name'],
             {
                 'metavar': 'role_name_or_id',
                 'help': 'Name or ID of the role to delete.'
             }),
        ])
    def enable(self):
        role: RoleInfo = self._client[self.app.pargs.role_name]
        self.app.print(self._client.enable(role))

    @ex(help='disable a role',
        arguments=[
            (['role_name'],
             {
                 'metavar': 'role_name_or_id',
                 'help': 'Name or ID of the role to delete.'
             }),
        ])
    def disable(self):
        role: RoleInfo = self._client[self.app.pargs.role_name]
        self.app.print(self._client.disable(role))



from banyan.api.base import ServiceBase
from banyan.model.policy import PolicyInfoOrName, PolicyAttachInfo
from banyan.model.service import ServiceInfo, Service, ServiceInfoOrName


class ServiceAPI(ServiceBase):
    class Meta:
        data_class = Service
        info_class = ServiceInfo
        arg_type = ServiceInfoOrName
        list_uri = '/registered_services'
        delete_uri = '/delete_registered_service'
        insert_uri = '/insert_registered_service'
        uri_param = 'ServiceID'
        obj_name = 'service'

    def enable(self, service: ServiceInfoOrName) -> str:
        service = self.find(service)
        json_response = self._client.api_request('POST',
                                                 '/enable_registered_service',
                                                 params={'ServiceID': service.id})
        return json_response['Message']

    def disable(self, service: ServiceInfoOrName) -> str:
        service = self.find(service)
        json_response = self._client.api_request('POST',
                                                 '/disable_registered_service',
                                                 params={'ServiceID': service.id})
        return json_response['Message']

    def attach(self, service: ServiceInfoOrName, policy: PolicyInfoOrName, enforcing: bool) -> PolicyAttachInfo:
        from banyan.api.policy import PolicyAPI
        return PolicyAPI(self._client).attach(policy, service, enforcing)

    def detach(self, service: ServiceInfoOrName, policy: PolicyInfoOrName) -> str:
        from banyan.api.policy import PolicyAPI
        return PolicyAPI(self._client).detach(policy, service)



from typing import List

from cement import Controller, ex

from banyan.api.shield import ShieldAPI
from banyan.controllers.base import Base
from banyan.core.exc import BanyanError
from banyan.model.netagent import Netagent
from banyan.model.shield import Shield


class ShieldController(Controller):
    class Meta:
        label = 'shield'
        stacked_type = 'nested'
        stacked_on = 'base'
        help = 'manage Banyan Shield clusters'

    @property
    def _client(self) -> ShieldAPI:
        return self.app.client.shields

    @ex(help='list shields')
    def list(self):
        shields: List[Shield] = self._client.list()
        results = list()
        headers = ['Name', 'ID', 'Address', 'SCEP Enabled', 'OTK Enabled', 'Status']
        for shield in shields:
            new_row = [shield.name, shield.id,
                       shield.public_addr, shield.scep_enabled, shield.otk_enabled, self._client.status(str(shield.id))]
            results.append(new_row)
        results.sort(key=lambda x: x[0])
        self.app.render(results, handler='tabulate', headers=headers, tablefmt='simple')

    @ex(help='get one-time key for netagent registration',
        arguments=[
            (['shield_name'],
             {
                 'metavar': 'shield_name_or_id',
                 'help': 'Name or ID of the shield.'
             }),
        ])
    def get_otk(self):
        shield_name = self.app.pargs.shield_name
        shield = self._client.find(shield_name)
        if self._client.status(shield_name) == 'INACTIVE':
            raise BanyanError(f'Shield {shield_name} status is INACTIVE')
        if not shield.otk_enabled:
            raise BanyanError(f'Shield {shield_name} is not enabled for OTK')
        self.app.print(str(shield.one_time_key))

    @ex(help='show detailed information about a shield',
        arguments=[
            (['shield_name'],
             {
                 'metavar': 'shield_name_or_id',
                 'help': 'Name or ID of the shield.'
             }),
        ])
    def get(self):
        shield_name = self.app.pargs.shield_name
        shield = self._client.find(shield_name)
        shield_json = Shield.Schema().dump(shield)
        # colorized_json = highlight(policy_json, lexers.JsonLexer(), formatters.Terminal256Formatter(style="default"))
        self.app.render(shield_json, handler='json', indent=2, sort_keys=True)

    @ex(help='list netagents connected to a shield',
        arguments=[
            (['shield_name'],
             {
                 'metavar': 'shield_name_or_id',
                 'help': 'Name or ID of the shield.'
             }),
        ])
    def list_netagents(self):
        shield_name = self.app.pargs.shield_name
        shield = self._client.find(shield_name)
        agents: List[Netagent] = self._client.config.netagent_map[str(shield.id)]
        results = list()
        headers = ['Site Name', 'Hostname', 'Public IP', 'Status', 'Last Activity']
        for agent in agents:
            new_row = [agent.name, agent.hostname, agent.public_ip_addr, agent.status,
                       agent.last_activity_at.strftime(Base.TABLE_DATE_FORMAT) if agent.last_activity_at else 'None']
            results.append(new_row)
        results.sort(key=lambda x: x[0] + x[1])
        self.app.render(results, handler='tabulate', headers=headers, tablefmt='simple')

    # TODO: implement for /shield_activity?num=1000&ShieldName=<name>
    @ex(help='show activity (stats, links, events) recorded by a shield')
    def list_activity(self):
        pass

    # TODO: implement /register_shield
    @ex(help='register a shield with the system')
    def create(self):
        pass

    # TODO: implement /delete_shield_cluster
    @ex(help='delete a shield')
    def delete(self):
        pass

    # TODO: implement /shield_events?StartTime=&EndTime=
    @ex(help='show security events recorded by a shield')
    def list_events(self):
        pass



from cement import App, TestApp, init_defaults
from cement.core.exc import CaughtSignal

from banyan.api import BanyanApiClient
from banyan.controllers.admin import AdminController
from banyan.controllers.base import Base
from banyan.controllers.device import DeviceController
from banyan.controllers.event import EventV1Controller
from banyan.controllers.netagent import NetagentController
from banyan.controllers.policy import PolicyController
from banyan.controllers.role import RoleController
from banyan.controllers.service import ServiceController
from banyan.controllers.shield import ShieldController
from banyan.controllers.user import UserController
from .core.exc import BanyanError

# configuration defaults
CONFIG = init_defaults('banyan')
CONFIG['banyan']['api_url'] = BanyanApiClient.DEFAULT_API_URL
CONFIG['banyan']['refresh_token'] = None


def extend_client(app: App) -> None:
    api_url = app.config.get('banyan', 'api_url')
    refresh_token = app.config.get('banyan', 'refresh_token')
    client = BanyanApiClient(api_url, refresh_token, debug=app.debug)
    app.extend('client', client)


class MyApp(App):
    """Banyan CLI primary application."""

    class Meta:
        label = 'banyan'

        # configuration defaults
        config_defaults = CONFIG

        # call sys.exit() on close
        exit_on_close = True

        # load additional framework extensions
        extensions = [
            # 'yaml',
            'colorlog',
            'json',
            'tabulate',
            # 'scrub',
            'print',
            # 'jinja2',
        ]

        hooks = [
            ('post_setup', extend_client)
        ]

        # configuration handler
        # config_handler = 'yaml'

        # configuration file suffix
        # config_file_suffix = '.yml'

        # set the log handler
        log_handler = 'colorlog'

        # set the output handler
        output_handler = 'tabulate'

        # register handlers
        handlers = [
            Base,
            ServiceController,
            RoleController,
            PolicyController,
            ShieldController,
            NetagentController,
            UserController,
            DeviceController,
            AdminController,
            EventV1Controller
        ]


class MyAppTest(TestApp, MyApp):
    """A sub-class of MyApp that is better suited for testing."""

    class Meta:
        label = 'banyan'


def main():
    with MyApp() as app:
        try:
            app.run()

        except AssertionError as e:
            print('AssertionError > %s' % e.args[0])
            app.exit_code = 1

            if app.debug is True:
                import traceback
                traceback.print_exc()

        except BanyanError as e:
            print('BanyanError > %s' % e.args[0])
            app.exit_code = 1

            if app.debug is True:
                import traceback
                traceback.print_exc()

        except CaughtSignal as e:
            # Default Cement signals are SIGINT and SIGTERM, exit 0 (non-error)
            print('\n%s' % e)
            app.exit_code = 0


if __name__ == '__main__':
    main()



from dataclasses import field
from datetime import datetime
from ipaddress import IPv4Address
from typing import List, Dict, ClassVar, Type
from uuid import UUID

import OpenSSL.crypto
from marshmallow import fields, Schema
from marshmallow_dataclass import dataclass
from semver import VersionInfo

from banyan.model import Resource
from banyan.model.netagent import Netagent


@dataclass
class Shield(Resource):
    org_id: UUID = field(metadata={"data_key": "OrgID"})
    shield_id: UUID = field(metadata={"data_key": "UUID"})
    shield_name: str = field(metadata={"data_key": "ShieldName"})
    group_type: str = field(metadata={"data_key": "GroupType"})
    cluster_mgr_type: str = field(metadata={"data_key": "ClusterMgrType"})
    cluster_mgr_ip: IPv4Address = field(metadata={"marshmallow_field": fields.String(data_key="ClusterMgrIP")})
    version: VersionInfo = field(metadata={"marshmallow_field": fields.String(data_key="ShieldVersion")})
    auto_upgrade: bool = field(metadata={"data_key": "AutoUpgrade"})
    one_time_key: UUID = field(metadata={"data_key": "OneTimeKey"})
    otk_expire_dt: datetime = field(metadata={"data_key": "OTKExpiryTime"})
    scep_enabled: bool = field(metadata={"data_key": "SCEPEnabled"})
    public_addr: str = field(metadata={"data_key": "PublicAddr"})
    ca_cert_str: str = field(metadata={"data_key": "CACert"})
    otk_enabled: bool = field(metadata={"data_key": "OTKEnabled"})
    ssh_ca_public_key: str = field(metadata={"data_key": "SSHCAPublicKey"})
    Schema: ClassVar[Type[Schema]] = Schema

    def load_cacert(self) -> OpenSSL.crypto.X509:
        cert = OpenSSL.crypto.load_certificate(OpenSSL.crypto.FILETYPE_PEM, self.ca_cert_str.encode('utf-8'))
        return cert

    @property
    def name(self) -> str:
        return self.shield_name

    @property
    def id(self) -> str:
        return str(self.shield_id)


@dataclass
class ShieldLastActivity:
    last_activity_time: datetime = field(metadata={"data_key": "InsertTime"})
    status: str = field(metadata={"data_key": "Status"})
    uuid: UUID = field(metadata={"data_key": "UUID"})
    Schema: ClassVar[Type[Schema]] = Schema


@dataclass
class ShieldConfig:
    shields: List[Shield] = field(default_factory=list, metadata={"data_key": "Configs"})
    last_activity_map: Dict[str, ShieldLastActivity] = field(default_factory=dict,
                                                             metadata={"data_key": "ShieldLastActivitiesMap"})
    netagent_map: Dict[str, List[Netagent]] = field(default_factory=dict,
                                                    metadata={"data_key": "NetagentHostInfoMap"})
    Schema: ClassVar[Type[Schema]] = Schema



from dataclasses import field
from datetime import datetime
from typing import List, ClassVar, Optional, Union
from uuid import UUID

from marshmallow import validate, Schema, pre_load
from marshmallow_dataclass import dataclass

from banyan.model import BanyanApiObject, InfoBase, NanoTimestampField


@dataclass
class Tags:
    TEMPLATE_USER = "USER"
    TEMPLATE_WORKLOAD = "WORKLOAD"
    TEMPLATE_CUSTOM = "CUSTOM"
    _TEMPLATE_VALUES = (TEMPLATE_USER, TEMPLATE_WORKLOAD, TEMPLATE_CUSTOM)
    template: str = field(metadata={"validate": validate.OneOf(_TEMPLATE_VALUES)})


@dataclass
class Metadata:
    name: str
    description: str
    tags: Tags = field(default_factory=Tags)


@dataclass
class Options:
    disable_tls_client_authentication: bool
    l7_protocol: Optional[str]
    mixed_users_and_workloads: Optional[bool]


@dataclass
class PolicyException:
    src_addr: List[str] = field(default_factory=list)
    tls_src_addr: List[str] = field(default_factory=list)  # Deprecated


@dataclass
class L7Access:
    PROTOCOL_HTTP = "http"
    PROTOCOL_KAFKA = "kafka"
    PROTOCOL_MYSQL = "mysql"
    RESOURCE_ALL = "*"
    ACTION_READ = "READ"
    ACTION_WRITE = "WRITE"
    ACTION_CREATE = "CREATE"
    ACTION_UPDATE = "UPDATE"
    ACTION_DELETE = "DELETE"
    ACTION_ALL = "*"
    _ACTION_VALUES = (ACTION_READ, ACTION_WRITE, ACTION_CREATE, ACTION_UPDATE, ACTION_DELETE, ACTION_ALL)
    action: Optional[str] = field(metadata={"validate": validate.OneOf(_ACTION_VALUES + ("",))})
    resources: List[str] = field(default_factory=list)
    actions: List[str] = field(default_factory=list)


@dataclass
class Conditions:
    TRUST_LEVEL_ALWAYS_DENY = "AlwaysDeny"
    TRUST_LEVEL_LOW = "Low"
    TRUST_LEVEL_MEDIUM = "Medium"
    TRUST_LEVEL_HIGH = "High"
    TRUST_LEVEL_ALWAYS_ALLOW = "AlwaysAllow"
    _TRUST_LEVEL_VALUES = (
        TRUST_LEVEL_ALWAYS_DENY, TRUST_LEVEL_LOW, TRUST_LEVEL_MEDIUM, TRUST_LEVEL_HIGH, TRUST_LEVEL_ALWAYS_ALLOW)
    start_time: Optional[datetime]
    end_time: Optional[datetime]
    trust_level: Optional[str] = field(metadata={"missing": TRUST_LEVEL_ALWAYS_ALLOW,
                                                 "validate": validate.OneOf(_TRUST_LEVEL_VALUES + ("",))})

    # noinspection PyUnusedLocal
    @pre_load
    def _replace_empty_dates(self, data, many, **kwargs):
        if "start_time" in data and data["start_time"] == "":
            del data["start_time"]
        if "end_time" in data and data["end_time"] == "":
            del data["end_time"]
        return data


@dataclass
class Rules:
    conditions: Conditions
    l7_access: Optional[List[L7Access]] = field(default_factory=list)


@dataclass
class Access:
    rules: Rules
    roles: List[str] = field(default_factory=list)


@dataclass
class Spec:
    options: Options
    exception: PolicyException
    access: List[Access] = field(default_factory=list)


@dataclass
class Policy(BanyanApiObject):
    metadata: Metadata
    spec: Spec
    type: str
    KIND = "BanyanPolicy"

    def __post_init__(self):
        self.kind = self.KIND

    @property
    def name(self):
        return self.metadata.name


@dataclass
class PolicyInfo(InfoBase):
    policy_id: UUID = field(metadata={'data_key': 'PolicyID'})
    policy_name: str = field(metadata={'data_key': 'PolicyName'})
    spec: str = field(metadata={'data_key': 'PolicySpec'})
    description: str = field(metadata={'data_key': 'Description'})
    version: int = field(metadata={'data_key': 'PolicyVersion'})
    Schema: ClassVar[Schema] = Schema

    @property
    def policy(self) -> Policy:
        return Policy.Schema().loads(self.spec)

    @property
    def name(self) -> str:
        return self.policy_name

    @property
    def id(self) -> str:
        return str(self.policy_id)


@dataclass
class PolicyAttachInfo:
    enabled: bool = field(metadata={'data_key': 'Enabled'})
    policy_id: UUID = field(metadata={'data_key': 'PolicyID'})
    service_id: str = field(metadata={'data_key': 'ServiceID'})
    attached_by: str = field(metadata={'data_key': 'AttachedBy'})
    detached_by: str = field(metadata={'data_key': 'DetachedBy'})
    attached_at: datetime = field(metadata={'marshmallow_field': NanoTimestampField(data_key='AttachedAt')})
    detached_at: datetime = field(metadata={'marshmallow_field': NanoTimestampField(data_key='DetachedAt')})
    Schema: ClassVar[Schema] = Schema


PolicyInfoOrName = Union[PolicyInfo, str]



from banyan.api.base import ServiceBase
from banyan.model import BanyanApiObject
from banyan.model.role import Role, RoleInfo, RoleInfoOrName


class RoleAPI(ServiceBase):
    class Meta:
        data_class = Role
        info_class = RoleInfo
        arg_type = RoleInfoOrName
        list_uri = '/security_roles'
        delete_uri = '/delete_security_role'
        insert_uri = '/insert_security_role'
        uri_param = 'RoleID'
        obj_name = 'role'

    def enable(self, role: RoleInfoOrName) -> str:
        role = self.find(role)
        json_response = self._client.api_request('POST',
                                                 '/enable_security_role',
                                                 params={'RoleID': str(role.id)})
        return json_response['Message']

    def disable(self, role: RoleInfoOrName) -> str:
        role = self.find(role)
        json_response = self._client.api_request('POST',
                                                 '/disable_security_role',
                                                 params={'RoleID': str(role.id)})
        return json_response['Message']

    def create(self, obj: BanyanApiObject) -> str:
        self._ensure_does_not_exist(obj.name)
        response_json = self._client.api_request('POST',
                                                 self.Meta.insert_uri,
                                                 json=obj.Schema().dump(obj))
        return response_json['RoleID']



from abc import ABC
from typing import Dict, List, Union

from banyan.core.exc import BanyanError
from banyan.model import InfoBase, BanyanApiObject, Resource, ResourceOrName

InfoObjectOrName = Union[InfoBase, str]


class ServiceBase(ABC):
    class Meta:
        data_class = BanyanApiObject
        info_class = Resource
        arg_type = ResourceOrName
        list_uri = '/undeclared'
        delete_uri = '/undeclared'
        insert_uri = '/undeclared'
        uri_param = 'ObjectID'
        obj_name = 'object'

    def __init__(self, client):
        self._client = client
        self._cache: List[Resource] = list()
        self._by_name: Dict[str, Resource] = dict()
        self._by_id: Dict[str, Resource] = dict()

    def list(self) -> list:
        response_json = self._client.api_request('GET', self.Meta.list_uri)
        data: List[Resource] = self.Meta.info_class.Schema().load(response_json, many=True)
        self._build_cache(data)
        return data

    def _build_cache(self, info: List[Resource]) -> None:
        self._cache = info
        self._by_name = {i.name.lower(): i for i in info}
        self._by_id = {str(i.id).lower(): i for i in info}

    def __getitem__(self, key: str):
        self._ensure_cache()
        try:
            return self._by_name[key.lower()]
        except KeyError:
            try:
                return self._by_id[key.lower()]
            except KeyError:
                raise BanyanError(f'{self.Meta.obj_name} name or ID does not exist: {key}')

    def exists(self, name: str) -> bool:
        self._ensure_cache()
        return name.lower() in self._by_name.keys() or name.lower() in self._by_id.keys()

    def _ensure_cache(self) -> None:
        if not self._cache:
            self.list()

    def _ensure_exists(self, name: str) -> None:
        if not self.exists(name):
            raise BanyanError(f'{self.Meta.obj_name} name does not exist: {name}')

    def _ensure_does_not_exist(self, name: str) -> None:
        if self.exists(name):
            raise BanyanError(f'{self.Meta.obj_name} name already exists: {name}')

    def find(self, obj: ResourceOrName):
        if isinstance(obj, Resource):
            self._ensure_exists(obj.name)
            return obj
        else:
            return self.__getitem__(obj)

    def create(self, obj: BanyanApiObject):
        if not self.Meta.insert_uri:
            raise BanyanError(f'{self.Meta.obj_name} API does not support creating objects')
        self._ensure_does_not_exist(obj.name)
        response_json = self._client.api_request('POST',
                                                 self.Meta.insert_uri,
                                                 json=obj.Schema().dump(obj))
        return self.Meta.info_class.Schema().load(response_json)

    def update(self, obj: BanyanApiObject):
        if not self.Meta.insert_uri:
            raise BanyanError(f'{self.Meta.obj_name} API does not support updating objects')
        self._ensure_exists(obj.name)
        response_json = self._client.api_request('POST',
                                                 self.Meta.insert_uri,
                                                 json=obj.Schema().dump(obj))
        return self.Meta.info_class.Schema().load(response_json)

    def delete(self, obj: ResourceOrName):
        if not self.Meta.delete_uri:
            raise BanyanError(f'{self.Meta.obj_name} API does not support deleting objects')
        obj = self.find(obj)
        json_response = self._client.api_request('DELETE',
                                                 self.Meta.delete_uri,
                                                 params={self.Meta.uri_param: str(obj.id)})
        return json_response['Message']



from datetime import timezone, datetime
from typing import List

from banyan.api.base import ServiceBase
from banyan.model.shield import Shield, ShieldConfig


class ShieldAPI(ServiceBase):
    class Meta:
        data_class = Shield
        info_class = Shield
        arg_type = str
        list_uri = '/shield_config'
        delete_uri = None
        insert_uri = None
        uri_param = None
        obj_name = 'shield'

    def __init__(self, client):
        self._config: ShieldConfig = ShieldConfig()
        super().__init__(client)

    def list(self) -> List[Shield]:
        response_json = self._client.api_request('GET', self.Meta.list_uri)
        self._config = ShieldConfig.Schema().load(response_json)
        self._build_cache(self._config.shields)
        return self._config.shields

    def active(self) -> List[Shield]:
        shields = self.list()
        # what defines a shield as "enabled"?
        return [x for x in shields if self.status(str(x.shield_id)) == 'REPORTING']

    def status(self, shield_name_or_id: str) -> str:
        shield: Shield = self.find(shield_name_or_id)
        activity = self.config.last_activity_map[shield.id]
        now = datetime.now(tz=timezone.utc)
        age = now - activity.last_activity_time
        return 'INACTIVE' if age.total_seconds() >= 900 else 'REPORTING'

    @property
    def config(self) -> ShieldConfig:
        self._ensure_cache()
        return self._config



from abc import ABC
from dataclasses import field
from datetime import datetime
from ipaddress import IPv4Interface
from typing import ClassVar, Type, Union

from marshmallow import fields, Schema
from marshmallow_dataclass import dataclass

from banyan.core.exc import BanyanError

API_VERSION = "rbac.banyanops.com/v1"


class NanoTimestampField(fields.DateTime):
    def _deserialize(self, value, attr, data, **kwargs):
        if not value:
            return None
        elif isinstance(value, int):
            return datetime.fromtimestamp(value / 1000000000)
        else:
            return super()._deserialize(value, attr, data, **kwargs)

    def _serialize(self, value: datetime, attr, obj, **kwargs):
        if not value:
            return 0
        elif isinstance(value, datetime):
            return int(value.timestamp() * 1000000000)
        else:
            return super()._serialize(value, attr, obj, **kwargs)


class IPv4InterfaceField(fields.Field):
    def _deserialize(self, value, attr, data, **kwargs):
        if not value:
            return None
        elif isinstance(value, str):
            return IPv4Interface(value)
        else:
            return super()._deserialize(value, attr, data, **kwargs)

    def _serialize(self, value, attr, data, **kwargs):
        if not value:
            return ""
        elif isinstance(value, IPv4Interface):
            return str(value)
        else:
            return super()._serialize(value, attr, data, **kwargs)


@dataclass
class BanyanApiObject:
    TYPE = "origin"
    apiVersion: str
    kind: str
    type: str
    Schema: ClassVar[Type[Schema]] = Schema

    def __post_init__(self):
        self.apiVersion = API_VERSION
        self.type = self.TYPE

    @property
    def name(self) -> str:
        raise NotImplementedError('not implemented in the base class')


@dataclass
class Resource(ABC):
    Schema: ClassVar[Type[Schema]] = Schema

    @property
    def name(self) -> str:
        raise BanyanError('not implemented')

    @property
    def id(self) -> str:
        raise BanyanError('not implemented')


ResourceOrName = Union[Resource, str]


@dataclass
class InfoBase(Resource, ABC):
    created_by: str = field(metadata={'data_key': 'CreatedBy'})
    created_at: datetime = field(metadata={'marshmallow_field': NanoTimestampField(data_key='CreatedAt')})
    last_updated_by: str = field(metadata={'data_key': 'LastUpdatedBy'})
    last_updated_at: datetime = field(metadata={'marshmallow_field': NanoTimestampField(data_key='LastUpdatedAt')})
    deleted_by: str = field(metadata={'data_key': 'DeletedBy'})
    deleted_at: datetime = field(metadata={'marshmallow_field': NanoTimestampField(data_key='DeletedAt')})
    spec: str
    description: str
    Schema: ClassVar[Type[Schema]] = Schema



from dataclasses import field
from datetime import datetime
from typing import List, ClassVar, Dict

from marshmallow import Schema
from marshmallow_dataclass import dataclass

from banyan.model import NanoTimestampField, Resource


@dataclass
class TrustFactorDetailV2:
    factor_name: str = field(metadata={'data_key': 'FactorName'})
    factor_value: str = field(metadata={'data_key': 'FactorValue'})
    is_active: bool = field(metadata={'data_key': 'IsActive'})
    inactive_reason: str = field(metadata={'data_key': 'InactiveReason'})
    relative_contribution: int = field(metadata={'data_key': 'RelativeContribution'})
    recommendation: str = field(metadata={'data_key': 'Recommendation'})


@dataclass
class TrustFactorDetailsV2:
    details: Dict[str, TrustFactorDetailV2] = field(default_factory=dict, metadata={'data_key': 'Details'})


@dataclass
class TrustFactorV2:
    id: str = field(metadata={'data_key': 'Id'})
    name: str = field(metadata={'data_key': 'Name'})
    value: str = field(metadata={'data_key': 'Value'})
    resource_id: str = field(metadata={'data_key': 'ResourceId'})
    reporting_epoch: int = field(metadata={'data_key': 'ReportingEpoch'})
    source: str = field(metadata={'data_key': 'Source'})
    source_type: str = field(metadata={'data_key': 'SourceType'})
    factor_id: str = field(metadata={'data_key': 'FactorId'})
    is_ml_factor: bool = field(metadata={'data_key': 'IsMLFactor'})
    input_features: Dict[str, str] = field(metadata={'data_key': 'InputFeatures'})
    scoring_algo: str = field(metadata={'data_key': 'ScoringAlgo'})


@dataclass
class TrustFactorsV2(Resource):
    score_id: str = field(metadata={'data_key': 'Id'})
    algorithm_id: str = field(metadata={'data_key': 'AlgorithmId'})
    resource_id: str = field(metadata={'data_key': 'ResourceId'})
    score: int = field(metadata={'data_key': 'Score'})
    score_type: str = field(metadata={'data_key': 'ScoreType'})
    timestamp: datetime = field(metadata={'marshmallow_field': NanoTimestampField(data_key='Timestamp')})
    explanation: TrustFactorDetailsV2 = field(metadata={'data_key': 'Explanation'})
    created_at: datetime = field(metadata={'marshmallow_field': NanoTimestampField(data_key='CreatedAt')})
    updated_at: datetime = field(metadata={'marshmallow_field': NanoTimestampField(data_key='UpdatedAt')})
    factors_updated_at: datetime = field(metadata={'marshmallow_field':
                                                   NanoTimestampField(data_key='FactorsUpdatedAt')})
    active_factors: List[TrustFactorV2] = field(default_factory=list, metadata={'data_key': 'ActiveFactors'})
    inactive_factors: List[TrustFactorV2] = field(default_factory=list, metadata={'data_key': 'InactiveFactors'})
    Schema: ClassVar[Schema] = Schema

    @property
    def name(self) -> str:
        return self.score_id

    @property
    def id(self) -> str:
        return self.score_id



from typing import List

from banyan.api.base import ServiceBase
from banyan.core.exc import BanyanError
from banyan.model.user_device import Device, TrustScore


class DeviceAPI(ServiceBase):
    UPDATE_FIELDS = ['architecture', 'model', 'platform', 'ownership', 'is_banned']

    class Meta:
        data_class = Device
        info_class = Device
        arg_type = str
        list_uri = '/devices'
        delete_uri = '/delete_device'
        insert_uri = '/mdm/update_device'
        uri_param = 'SerialNumber'
        obj_name = 'device'

    def create(self, obj: Device):
        raise BanyanError('devices cannot be created via the API')

    def update(self, obj: Device) -> str:
        self._ensure_exists(obj.serial_number)
        response_json = self._client.api_request('POST',
                                                 self.Meta.insert_uri,
                                                 params={self.Meta.uri_param: obj.serial_number},
                                                 json=obj.Schema(only=DeviceAPI.UPDATE_FIELDS).dump(obj))
        return response_json['Message']

    def ban(self, obj: Device) -> str:
        obj.is_banned = True
        return self.update(obj)

    def unban(self, obj: Device) -> str:
        obj.is_banned = False
        return self.update(obj)

    def set_max_trustlevel(self, obj: Device, max_level: str, reason: str, ext_source: str) -> str:
        self._ensure_exists(obj.serial_number)
        response_json = self._client.api_request('POST',
                                                 '/set_max_trust_level',
                                                 params={self.Meta.uri_param: obj.serial_number},
                                                 json={
                                                     'Level': max_level,
                                                     'Reason': reason,
                                                     'ExtSource': ext_source
                                                 })
        return TrustScore.Schema().load(response_json)

    def get_trustscores(self, obj: Device) -> List[TrustScore]:
        self._ensure_exists(obj.serial_number)
        response_json = self._client.api_request('GET',
                                                 '/trustscore',
                                                 params={self.Meta.uri_param: obj.serial_number})
        return TrustScore.Schema().load(response_json, many=True)



from typing import List

from cement import Controller, ex

from banyan.api.service import ServiceAPI
from banyan.controllers.base import Base
from banyan.model.service import ServiceInfo, Service


class ServiceController(Controller):
    class Meta:
        label = 'service'
        stacked_type = 'nested'
        stacked_on = 'base'
        help = 'manage web and TCP services and workloads'

    @property
    def _client(self) -> ServiceAPI:
        return self.app.client.services

    @ex(help='list registered services')
    def list(self):
        services: List[ServiceInfo] = self._client.list()
        results = list()
        headers = ['Name', 'ID', 'Type', 'Enabled', 'Created', 'Last Updated']
        for service in services:
            app_type = service.service.metadata.tags.service_app_type
            new_row = [service.service_name, service.service_id, app_type, service.enabled,
                       service.created_at.strftime(Base.TABLE_DATE_FORMAT),
                       service.last_updated_at.strftime(Base.TABLE_DATE_FORMAT)]
            results.append(new_row)
        results.sort(key=lambda x: x[0])
        self.app.render(results, handler='tabulate', headers=headers, tablefmt='simple')

    @ex(help='show the definition of a registered service',
        arguments=[
            (['service_name'],
             {
                 'metavar': 'service_name_or_id',
                 'help': 'Name or ID of the service to display.'
             }),
        ])
    def get(self):
        info: ServiceInfo = self._client[self.app.pargs.service_name]
        service_json = Service.Schema().dump(info.service)
        # colorized_json = highlight(service_json, lexers.JsonLexer(), formatters.Terminal256Formatter(style="default"))
        self.app.render(service_json, handler='json', indent=2, sort_keys=True)

    @ex(help='create a new service from a JSON specification',
        arguments=[
            (['service_spec'],
             {
                 'help': 'JSON blob describing the new service to be created, or a filename '
                         'containing JSON prefixed by "@" (example: @service.json).'
             }),
        ])
    def create(self):
        spec = Base.get_json_input(self.app.pargs.service_spec)
        service = Service.Schema().load(spec)
        info = self._client.create(service)
        self.app.render(ServiceInfo.Schema().dump(info), handler='json', indent=2, sort_keys=True)

    @ex(help='update an existing service from a JSON specification',
        arguments=[
            (['service_spec'],
             {
                 'help': 'JSON blob describing the service to be updated, or a filename '
                         'containing JSON prefixed by "@" (example: @service.json).'
             }),
        ])
    def update(self):
        spec = Base.get_json_input(self.app.pargs.service_spec)
        service = Service.Schema().load(spec)
        info = self._client.update(service)
        self.app.render(ServiceInfo.Schema().dump(info), handler='json', indent=2, sort_keys=True)

    @ex(help='delete a service',
        arguments=[
            (['service_name'],
             {
                 'metavar': 'service_name_or_id',
                 'help': 'Name or ID of the service to delete.'
             }),
        ])
    def delete(self):
        service: ServiceInfo = self._client[self.app.pargs.service_name]
        self.app.print(self._client.delete(service))

    @ex(help='enable a service',
        arguments=[
            (['service_name'],
             {
                 'metavar': 'service_name_or_id',
                 'help': 'Name or ID of the service to enable.'
             }),
        ])
    def enable(self):
        service: ServiceInfo = self._client[self.app.pargs.service_name]
        self.app.print(self._client.enable(service))

    @ex(help='disable a service',
        arguments=[
            (['service_name'],
             {
                 'metavar': 'service_name_or_id',
                 'help': 'Name or ID of the service to disable.'
             }),
        ])
    def disable(self):
        service: ServiceInfo = self._client[self.app.pargs.service_name]
        self.app.print(self._client.disable(service))

    @ex(help='attach a policy to a service',
        arguments=[
            (['service_name'],
             {
                 'metavar': 'service_name_or_id',
                 'help': 'Name or ID of the service to attach a policy to.',
             }),
            (['policy_name'],
             {
                 'metavar': 'policy_name_or_id',
                 'help': 'Name or ID of the policy to attach to the service.',
             }),
            (['--permissive'],
             {
                 'action': 'store_false',
                 'dest': 'enforcing',
                 'help': 'Set the policy to permissive mode (allow all traffic and log any unauthorized access).',
             }),
            (['--enforcing'], {
                'action': 'store_true',
                'dest': 'enforcing',
                'default': True,
                'help': 'Set the policy to enforcing mode (deny unauthorized access).',
            }),
        ])
    def attach_policy(self):
        result = self._client.attach(self.app.pargs.service_name, self.app.pargs.policy_name, self.app.pargs.enforcing)
        mode = 'ENFORCING' if result.enabled else 'PERMISSIVE'
        self.app.print(f'Policy {result.policy_id} attached to service {result.service_id} in {mode} mode.')

    @ex(help='detach the active policy from a service',
        arguments=[
            (['service_name'],
             {
                 'metavar': 'service_name_or_id',
                 'help': 'Name or ID of the service to attach a policy to.',
             }),
            (['policy_name'],
             {
                 'metavar': 'policy_name_or_id',
                 'help': 'Name or ID of the policy to attach to the service.',
             }),
        ])
    def detach_policy(self):
        self.app.print(self._client.detach(self.app.pargs.service_name, self.app.pargs.policy_name))

    @ex(help='run sanity checks on a service')
    def test(self):
        pass



from typing import List

from banyan.api.base import ServiceBase
from banyan.model.netagent import Netagent


class NetagentAPI(ServiceBase):
    class Meta:
        data_class = Netagent
        info_class = Netagent
        arg_type = str
        list_uri = '/netagents'
        delete_uri = None
        insert_uri = None
        uri_param = None
        obj_name = 'netagent'

    def active(self) -> List[Netagent]:
        agents: List[Netagent] = self.list()
        return [x for x in agents if x.status == 'REPORTING']



from typing import List

from cement import Controller, ex

from banyan.api import DeviceAPI
from banyan.controllers.base import Base
from banyan.model.user_device import Device, TrustDataV1, TrustScore


class DeviceController(Controller):
    class Meta:
        label = 'device'
        stacked_type = 'nested'
        stacked_on = 'base'
        help = 'manage devices'

    @property
    def _client(self) -> DeviceAPI:
        return self.app.client.devices

    @ex(help='list devices')
    def list(self):
        devices: List[Device] = self._client.list()
        results = list()
        headers = ['Friendly Name', 'Serial Number', 'Platform', 'Ownership', 'Trust Score', 'Banned', 'Last Login']
        for device in devices:
            new_row = [device.device_friendly_name, device.serial_number, device.platform,
                       device.ownership, device.trust_data.level, device.is_banned,
                       device.last_login.strftime(Base.TABLE_DATE_FORMAT)]
            results.append(new_row)
        results.sort(key=lambda x: x[0])
        self.app.render(results, handler='tabulate', headers=headers, tablefmt='simple')

    @ex(help='show detailed information about a device',
        arguments=[
            (['serial_number'],
             {
                 'help': 'Serial number of the device.'
             }),
        ])
    def get(self):
        serial_number = self.app.pargs.serial_number
        device = self._client.find(serial_number)
        device_json = Device.Schema().dump(device)
        # colorized_json = highlight(policy_json, lexers.JsonLexer(), formatters.Terminal256Formatter(style="default"))
        self.app.render(device_json, handler='json', indent=2, sort_keys=True)

    # TODO: implement /mdm/update_device?SerialNumber=X
    @ex(help='update device status')
    def update(self):
        pass

    # TODO: implement /delete_device?Email=X&SerialNumber=Y
    @ex(help='delete a device')
    def delete(self):
        pass

    # TODO: implement /device/unregister
    @ex(help='unregister a device')
    def unregister(self):
        pass

    # TODO: implement /devices/stats
    @ex(help='display statistics about registered devices')
    def summary(self):
        pass

    @ex(help='ban a device (prevent it from accessing network resources)',
        arguments=[
            (['serial_number'],
             {
                 'help': 'Serial number of the device.'
             }),
        ])
    def ban(self):
        serial_number = self.app.pargs.serial_number
        device = self._client.find(serial_number)
        self.app.print(self._client.ban(device))

    @ex(help='remove the banned status from a device',
        arguments=[
            (['serial_number'],
             {
                 'help': 'Serial number of the device.'
             }),
        ])
    def unban(self):
        serial_number = self.app.pargs.serial_number
        device = self._client.find(serial_number)
        self.app.print(self._client.unban(device))

    @ex(help='set the maximum trust level for a device',
        arguments=[
            (['serial_number'],
             {
                 'help': 'Serial number of the device.'
             }),
            (['--trust-level'],
             {
                 'choices': TrustDataV1.TRUST_VALUES,
                 'required': True,
                 'help': f'Maximum trust level for this device. Must be one of: {TrustDataV1.TRUST_VALUES}.'
             }
             ),
            (['--reason'],
             {
                 'required': True,
                 'help': 'Explanation to be displayed in console and to the end user.'
             }),
            (['--ext-source'],
             {
                 'required': True,
                 'help': 'Name of the external data source (e.g. CarbonBlack, CrowdStrike, etc).'
             }),
        ])
    def set_max_trust(self):
        serial_number = self.app.pargs.serial_number
        device = self._client.find(serial_number)
        trust = self._client.set_max_trustlevel(device, self.app.pargs.trust_level,
                                                self.app.pargs.reason, self.app.pargs.ext_source)
        trust_json = TrustScore.Schema().dump(trust)
        self.app.render(trust_json, handler='json', indent=2, sort_keys=True)

    @ex(help='show details about the device\'s trust score calculation',
        arguments=[
            (['serial_number'],
             {
                 'help': 'Serial number of the device.'
             }),
        ])
    def get_trust(self):
        serial_number = self.app.pargs.serial_number
        device = self._client.find(serial_number)
        trust = self._client.get_trustscores(device)
        trust_json = TrustScore.Schema().dump(trust, many=True)
        # colorized_json = highlight(policy_json, lexers.JsonLexer(), formatters.Terminal256Formatter(style="default"))
        self.app.render(trust_json, handler='json', indent=2, sort_keys=True)



from typing import List

from banyan.api.base import ServiceBase
from banyan.model.user_device import User, TrustScore


class UserAPI(ServiceBase):
    class Meta:
        data_class = User
        info_class = User
        arg_type = str
        list_uri = '/endusers'
        delete_uri = None
        insert_uri = None
        uri_param = 'Email'
        obj_name = 'user'

    def set_max_trustlevel(self, obj: User, max_level: str, reason: str, ext_source: str) -> str:
        self._ensure_exists(obj.email)
        response_json = self._client.api_request('POST',
                                                 '/set_max_trust_level',
                                                 params={self.Meta.uri_param: obj.email},
                                                 json={
                                                     'Level': max_level,
                                                     'Reason': reason,
                                                     'ExtSource': ext_source
                                                 })
        return TrustScore.Schema().load(response_json)

    def get_trustscores(self, obj: User) -> List[TrustScore]:
        self._ensure_exists(obj.email)
        response_json = self._client.api_request('GET',
                                                 '/trustscore',
                                                 params={self.Meta.uri_param: obj.email})
        return TrustScore.Schema().load(response_json, many=True)
